import inspect
from collections import OrderedDict

from .dot_dict import DotDict


def get_args(func):
    """
    获取函数的参数列表（带参数类型）
    :param func:
    :return:
    """
    parameters = inspect.signature(func).parameters

    args = OrderedDict()

    for p in parameters.keys():
        spec = DotDict()
        # 类型
        annotation = parameters.get(p).annotation
        default = parameters.get(p).default
        if default != inspect._empty:
            spec['default'] = default

        # 有默认值时，若未指定类型，则使用默认值的类型
        if annotation == inspect._empty:
            if default is not None and default != inspect._empty:
                spec['annotation'] = type(default)
        else:
            spec['annotation'] = annotation

        args[p] = spec

    return args



import re


class DotDict(dict):
    """
    定义一个允许通过点号访问的dict
    """

    def __init__(self, *args, **kwargs):
        """
        初始化
        :param args:
        :param kwargs:
        """
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self

    @staticmethod
    def parse(obj, handle_key=True):
        """
        把 dict 转换成可以通过 . 号访问的 DotDict
        :param obj: dict,list
        :param handle_key: 是否处理特殊 key
        :return:
        """
        dot_dict = DotDict()

        # 处理 dict
        if isinstance(obj, dict):
            for (key, value) in obj.items():
                if handle_key:
                    # 处理特殊的key
                    zhPattern = re.compile(u'[\u4e00-\u9fa5]+')
                    match = zhPattern.search(key)
                    if not match:
                        key = re.sub(r'[^0-9a-zA-Z_]', '_', key)
                        # 如果第一个字符不是标识符 就在前面添加一个下划线
                        if not re.match(r'^[a-zA-Z_]', key):
                            key = '_' + key
                setattr(dot_dict, key, DotDict.__parse_value(value, handle_key))
            return dot_dict

        # 处理list
        return list(map(lambda item: DotDict.parse(item), obj))

    @staticmethod
    def __parse_value(value, handle_key=True):
        # 处理列表类型
        if isinstance(value, list):
            return list(map(lambda item: DotDict.__parse_value(item, handle_key), value))

        # 处理字典类型
        if isinstance(value, dict):
            return DotDict.parse(value, handle_key)

        # 普通类型
        return value



import os

from django.conf import settings
from django.http import HttpResponse

from .util import logger

# 注册的路由中间件列表
MIDDLEWARE_INSTANCE_LIST = []


def add_middleware(middleware_name):
    """
    添加路由中间件
    :param middleware_name:中间件的模块完全限定名称
    :return:
    """

    temp = middleware_name.split('.')

    class_name = temp.pop()

    module_name = '.'.join(temp)

    module_path = os.path.join(settings.BASE_DIR, module_name.replace('.', os.path.sep))

    if not os.path.exists(module_path + '.py'):
        # 是目录 (package)，尝试使用默认的
        if os.path.exists(module_path):
            module_path = os.path.join(module_path, '__init__.py')
            module_name += '.__init__'

        if not os.path.exists(module_path):
            logger.error('Middleware for restful-dj is not found: %s' % middleware_name)
            return

    module = __import__(module_name, fromlist=True)

    middleware = getattr(module, class_name)

    if middleware not in MIDDLEWARE_INSTANCE_LIST:
        MIDDLEWARE_INSTANCE_LIST.append(middleware())

    return True


class MiddlewareBase:
    """
    路由中间件基类
    """

    def __init__(self):
        self.session = {}

    def request_process(self, request):
        """
        对 request 对象进行预处理。一般用于请求的数据的解码
        :param request:
        :return: 返回 HttpResponse 以终止请求
        """
        pass

    def response_process(self, request, response):
        """
        对 response 数据进行预处理。一般用于响应的数据的编码
        :param request:
        :param response:
        :return: 应该始终返回一个  HttpResponse
        """
        return response

    def check_login_status(self, request, meta):
        """
        检查用户的登录状态，使用时请覆写此方法
        :param request:
        :param meta:
        :return: True|False|HttpResponse 已经登录时返回 True，否则返回 False，HttpResponse 响应
        """
        return True

    def check_user_permission(self, request, meta):
        """
        检查用户是否有权限访问此路由，使用时请覆写此方法
        :param request:
        :param meta:
        :return: True|False|HttpResponse 已经登录时返回 True，否则返回 False，HttpResponse 响应
        """
        return True

    def check_params(self, request, meta):
        """
        在调用路由函数前，对参数进行处理，使用时请覆写此方法
        :param request:
        :param meta:
        :return: 返回 HttpResponse 以终止请求
        """
        pass

    def process_return_value(self, request, meta, data):
        """
        在路由函数调用后，对其返回值进行处理
        :param request:
        :param meta:
        :param data:
        :return:
        """
        return data

class MiddlewareManager:
    """
    路由中间件管理器
    """

    def __init__(self):
        # 处理函数的id
        self.id = None
        # 处理函数的名称
        self.handler = None
        # 路由模块名称
        self.module = None
        # 路由名称
        self.name = None
        # 是否需要用户权限
        self.permission_required = True
        # HTTP请求对象
        self.request = None

    def invoke(self):
        # 对 request 进行预处理
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'request_process'):
                continue
            result = middleware.request_process(self.request)
            if isinstance(result, HttpResponse):
                return result

        # 不需要登录，也不需要检查用户权限
        if not self.permission_required:
            return True

        # 需要检查登录状态
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'check_login_status'):
                continue

            result = middleware.check_login_status(self.request, {
                'id': self.id,
                'handler': self.handler,
                'module': self.module,
                'name': self.name
            })

            # 没有返回结果
            if result is None:
                continue

            # 如果路由中间件有返回结果
            if isinstance(result, HttpResponse):
                return result

            # 只要不返回 true 都表示不能继续了
            if result is False:
                return result

        # 需要检查用户权限
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if self.permission_required and hasattr(middleware, 'check_user_permission'):
                result = middleware.check_user_permission(self.request, {
                    'id': self.id,
                    'handler': self.handler,
                    'module': self.module,
                    'name': self.name
                })

                # 没有返回结果
                if result is None:
                    continue

                # 如果路由中间件有返回结果
                if isinstance(result, HttpResponse):
                    return result

                # 只要不返回 true 都表示不能继续了
                if result is not True:
                    return result
        return True

    def params(self):
        """
        在调用路由函数前，对参数进行处理
        :param request:
        :return:
        """
        # 对 response 进行处理
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'check_params'):
                continue
            result = middleware.check_params(self.request, {
                'id': self.id,
                'handler': self.handler,
                'module': self.module,
                'name': self.name
            })

            # 没有返回结果
            if result is None:
                continue

            # 如果路由中间件有返回结果
            if isinstance(result, HttpResponse):
                return result

    def process_return(self, data):
        """
        在路由函数调用后，对其返回值进行处理
        :param data:
        :return:
        """
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'process_return_value'):
                continue
            data = middleware.process_return_value(self.request, {
                'id': self.id,
                'handler': self.handler,
                'module': self.module,
                'name': self.name
            }, data)
        return data

    def end(self, response):
        """
        在响应前，对响应的数据进行处理
        :param response:
        :return:
        """
        # 对 response 进行处理
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'response_process'):
                continue
            response = middleware.response_process(self.request, response)

        return response



from django.conf import settings

from .util import logger

from .middleware import add_middleware

APP_CONFIG_KEY = 'RESTFUL_DJ'
APP_CONFIG_ROUTE = 'routes'
APP_CONFIG_MIDDLEWARE = 'middleware'
APP_CONFIG_LOGGER = 'logger'

if not hasattr(settings, APP_CONFIG_KEY):
    logger.error('config item not found in settings.py: %s' % APP_CONFIG_KEY)

# 已注册APP集合
CONFIG_ROOT: dict = getattr(settings, APP_CONFIG_KEY)

if APP_CONFIG_ROUTE not in CONFIG_ROOT:
    logger.error('config item not found in settings.py!%s: %s' % (APP_CONFIG_KEY, APP_CONFIG_ROUTE))

CONFIG_ROUTE: list = []
for _ in CONFIG_ROOT[APP_CONFIG_ROUTE].keys():
    CONFIG_ROUTE.append(_)
CONFIG_ROUTE = sorted(CONFIG_ROUTE, key=lambda i: len(i), reverse=True)

if APP_CONFIG_MIDDLEWARE in CONFIG_ROOT:
    for middleware in CONFIG_ROOT[APP_CONFIG_MIDDLEWARE]:
        add_middleware(middleware)

if APP_CONFIG_LOGGER in CONFIG_ROOT:
    custom_logger_name = CONFIG_ROOT[APP_CONFIG_MIDDLEWARE]
    logger.set_logger(__import__(custom_logger_name, fromlist=True))



import json
from functools import wraps

from django.http import HttpResponse, JsonResponse, HttpResponseBadRequest

from .middleware import MiddlewareManager
from .util.dot_dict import DotDict

from .util import logger


def route(module=None, name=None, permission=True, ajax=True, referer=None, **kwargs):
    """
    用于控制路由的访问权限，路由均需要添加此装饰器，若未添加，则不可访问
    :param module: str 路由所属模块，一般在查询权限时会用到
    :param name: str 路由名称，一般在查询权限时会用到
    :param permission: bool 访问此地址是否需要检查用户权限,由数据库实际值控制,在这里只起到做入库作用
    :param ajax: bool 是否仅允许ajax请求,由数据库实际值控制,在这里只起到做入库作用
    :param referer: list|str 允许的来源页,由数据库实际值控制,在这里只起到做入库作用
    用法：
    @route('用户管理', '编辑用户', permission=True)
    def get(req):
        pass
    """

    def invoke_route(func):
        @wraps(func)
        def caller(request, args):
            func_name = func.__name__

            mgr = MiddlewareManager()
            mgr.id = '{0}_{1}'.format(func.__module__.replace('_', '__').replace('.', '_'), func_name)
            mgr.handler = func_name
            mgr.module = module
            mgr.name = name
            mgr.permission_required = permission
            mgr.request = request

            # 调用中间件以检查登录状态以及用户权限
            result = mgr.invoke()

            # 返回了 HttpResponse ， 直接返回此对象
            if isinstance(result, HttpResponse):
                return mgr.end(result)

            # 返回了 False，表示未授权访问
            if result is False:
                return mgr.end(HttpResponseUnauthorized())

            # 处理请求中的json参数
            # 处理后可能会在 request 上添加一个 json 的项，此项存放着json格式的 body 内容
            _process_json_params(request)

            result = mgr.params()

            # 返回了 HttpResponse ， 直接返回此对象
            if isinstance(result, HttpResponse):
                return mgr.end(result)

            # 调用路由处理函数
            arg_len = len(args)
            method = request.method.lower()
            if arg_len == 0:
                return mgr.end(_wrap_http_response(mgr, func()))

            if arg_len == 1:
                return mgr.end(_wrap_http_response(mgr, func(request)))

            # 多个参数，自动从 queryString, POST 或 json 中获取
            # 匹配参数

            actual_args = [request]

            import inspect
            signature = inspect.signature(func)

            # 规定：第一个参数只能是  request，所以此处直接跳过第一个参数
            position = 0
            for arg_name in args.keys():
                position += 1
                if position == 1:
                    continue

                arg_spec = args.get(arg_name)

                if method in ['delete', 'get']:
                    use_default, arg_value = _get_value(request.G, arg_name, arg_spec, signature)
                else:
                    use_default, arg_value = _get_value(request.P, arg_name, arg_spec, signature, request.B)

                # 未找到参数
                if use_default is None:
                    return mgr.end(HttpResponseBadRequest('Parameter "%s" is required' % arg_name))

                # 使用默认值
                if use_default is True:
                    actual_args.append(arg_value)
                    continue

                # 未指定类型
                if 'annotation' not in arg_spec:
                    actual_args.append(arg_value)
                    continue

                # 检查类型是否一致 #

                # 类型一致，直接使用
                if isinstance(arg_value, arg_spec.annotation):
                    actual_args.append(arg_value)
                    continue

                # 类型不一致，尝试转换类型
                # 转换失败时，会抛出异常
                try:
                    actual_args.append(arg_spec.annotation(arg_value))
                except Exception as e:
                    msg = 'Parameter type of "%s" mismatch, signature: %s' % (arg_name, _get_signature(signature))
                    logger.warning(msg)
                    return mgr.end(HttpResponseBadRequest(msg))

            result = func(*actual_args)

            return mgr.end(_wrap_http_response(mgr, result))

        return caller

    return invoke_route


def _process_json_params(request):
    """
    参数处理
    :return:
    """
    request.B = DotDict()
    request.G = DotDict()
    request.P = DotDict()

    if request.content_type != 'application/json':
        request.G = DotDict.parse(request.GET.dict())
        request.P = DotDict.parse(request.POST.dict())
        return

    # 如果请求是json类型，就先处理一下

    body = request.body

    if body == '' or body is None:
        return

    try:
        if isinstance(body, bytes) or isinstance(body, str):
            request.B = DotDict.parse(json.loads(body))
        elif isinstance(body, dict) or isinstance(body, list):
            request.B = DotDict.parse(body)
    except Exception as e:
        logger.warning('Deserialize request body fail: %s' % str(e))


def _get_signature(signature):
    # 第一个总是  request,去掉
    return '(%s)' % ','.join(str(signature).strip('(').strip(')').split(',')[1:]).strip()


def _parameter_is_missing(signature, arg_name):
    logger.error('Missing required parameter "%s", signature: %s' % (arg_name, _get_signature(signature)))


def _get_value(data: dict, name: str, arg_spec: DotDict, signature, backup: dict = None):
    """

    :param data:
    :param name:
    :param arg_spec:
    :param signature:
    :return: True 表示使用默认值 False 表示未使用默认值 None 表示无值
    """
    # 内容变量时，移除末尾的 _ 符号
    # 内容变量时，移除末尾的 _ 符号
    inner_name = name.rstrip('_')
    if name in data:
        return False, data[name]

    if inner_name in data:
        return False, data[inner_name]

    if backup is not None:
        if name in backup:
            return False, backup[name]

        if inner_name in backup:
            return False, backup[inner_name]

    # 尝试使用默认值
    if 'default' in arg_spec:
        return True, arg_spec.default

    # 缺少无默认值的参数
    _parameter_is_missing(signature, name)
    return None, None


def _wrap_http_response(mgr, data):
    """
    将数据包装成 HttpResponse 返回
    :param data:
    :return:
    """

    # 处理返回函数
    data = mgr.process_return(data)

    if data is None:
        return HttpResponse()

    if isinstance(data, HttpResponse):
        return data

    if isinstance(data, bool):
        return HttpResponse('true' if bool else 'false')

    if isinstance(data, (dict, list, set, tuple, DotDict)):
        return JsonResponse(data, safe=False)

    if isinstance(data, str):
        return HttpResponse(data.encode())

    if isinstance(data, bytes):
        return HttpResponse(data)

    return HttpResponse(str(data).encode())


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401



import inspect
import os

from django.conf import settings
from django.http import HttpResponseNotFound, HttpResponseServerError, HttpRequest, HttpResponse
from .setting import CONFIG_ROUTE, CONFIG_ROOT, APP_CONFIG_ROUTE
from .util import collector, func_util

from .util import logger

# 包名称
from .util.dot_dict import DotDict

NAME = 'restful_dj'

# 函数缓存，减少 inspect 反射调用次数
ENTRY_CACHE = {}

_BEFORE_DISPATCH_HANDLER = None

# 线上模式时，使用固定路由
PRODUCTION_ROUTES = {}

if not settings.DEBUG:
    for _route in collector.collect():
        func = getattr(__import__(_route['pkg'], fromlist=True), _route['handler'])
        args = func_util.get_args(func)
        rid = '%s#%s' % (_route['path'], _route['method'])
        PRODUCTION_ROUTES[rid] = DotDict.parse({
            'func': func,
            'args': args
        })


def set_before_dispatch_handler(handler):
    """
    设置请求分发前的处理函数
    :param handler:
    :return:
    """
    global _BEFORE_DISPATCH_HANDLER
    _BEFORE_DISPATCH_HANDLER = handler


def dispatch(request, entry, name=''):
    """
    REST-ful 路由分发入口
    :param request: 请求
    :param entry: 入口文件，包名使用 . 符号分隔
    :param name='' 指定的函数名称
    :return:
    """
    if _BEFORE_DISPATCH_HANDLER is not None:
        entry, name = _BEFORE_DISPATCH_HANDLER(request, entry, name)

    if not settings.DEBUG:
        return _route_for_production(request, entry, name)

    router = Router(request, entry, name)
    check_result = router.check()
    if isinstance(check_result, HttpResponse):
        return check_result
    return router.route()


def _route_for_production(request, entry, name):
    method = request.method.lower()
    try:
        if name == '':
            route = PRODUCTION_ROUTES['%s#%s' % (entry, method)]
        else:
            route = PRODUCTION_ROUTES['%s/%s#%s' % (entry, name, method)]
    except Exception:
        return HttpResponseNotFound()

    return _invoke_handler(request, route.func, route.args)


def _invoke_handler(request, func, args):
    try:
        return func(request, args)
    except Exception as e:
        message = 'Router invoke exception'
        logger.error(message, e)
        return HttpResponseServerError('%s: %s' % (message, str(e)))


class Router:
    def __init__(self, request: HttpRequest, entry: str, name: str):
        self.request = request
        self.entry = entry
        method = request.method.lower()
        self.method = method

        # 如果指定了名称，那么就加上
        # 如：name = 'detail'
        #   func_name = get_detail
        if name:
            func_name = '%s_%s' % (method, name.lower())
        else:
            func_name = method
        self.func_name = func_name

        # 处理映射
        # 对应的模块(文件路径）
        self.module_name = self.get_route_map(entry)

        self.fullname = ''

    def check(self):
        module_name = self.module_name

        if module_name is None:
            logger.warning('Cannot find route map in RESTFUL_DJ.routes: %s' % self.entry)
            return HttpResponseNotFound()

        # 如果 module_name 是目录，那么就查找 __init__.py 是否存在
        abs_path = os.path.join(settings.BASE_DIR, module_name.replace('.', os.path.sep))
        if os.path.isdir(abs_path):
            logger.info('Entry "%s" is package, auto load module "__init__.py"' % module_name)
            module_name = '%s.%s' % (module_name, '__init__')
        elif not os.path.exists('%s.py' % abs_path):
            return HttpResponseNotFound()

        self.module_name = module_name

        # 完全限定名称
        self.fullname = '%s.%s' % (module_name, self.func_name)

    def route(self):
        try:
            func_define = self.get_func_define()
        except Exception as e:
            message = 'Load entry "%s" failed' % self.module_name
            logger.error(message, e)
            return HttpResponseNotFound()

        # 如果 func_define 为 False ，那就表示此函数不存在
        if func_define is False:
            message = 'Route "%s.%s" not found' % (self.module_name, self.func_name)
            logger.error(message)
            return HttpResponseNotFound()

        if func_define is HttpResponse:
            return func_define

        return _invoke_handler(self.request, func_define.func, func_define.args)

    def get_func_define(self):
        fullname = self.fullname
        func_name = self.func_name
        module_name = self.module_name

        # 缓存中有这个函数
        if fullname in ENTRY_CACHE.keys():
            return ENTRY_CACHE[fullname]

        # 缓存中没有这个函数，去模块中查找
        # ---------------

        try:
            # 如果不加上fromlist=True,只会导入目录
            # noinspection PyTypeChecker
            # __import__ 自带缓存
            entry_define = __import__(module_name, fromlist=True)
        except Exception as e:
            message = 'Load module "%s" failed' % module_name
            logger.error(message, e)
            return HttpResponseNotFound()

        # 模块中也没有这个函数
        if not hasattr(entry_define, func_name):
            # 函数不存在，更新缓存
            ENTRY_CACHE[func_name] = False
            return False

        # 模块中有这个函数
        # 通过反射从模块加载函数
        func = getattr(entry_define, func_name)
        if not self.is_valid_route(func):
            msg = 'Decorator "@route" not found on function "%s", did you forgot it ?' % fullname
            logger.warning(msg)
            # 没有配置装饰器@route，则认为函数不可访问，更新缓存
            ENTRY_CACHE[func_name] = False
            return False

        ENTRY_CACHE[fullname] = DotDict.parse({
            'func': func,
            # 该函数的参数列表
            # 'name': {
            #     'annotation': '类型', 当未指定类型时，无此项
            #     'default': '默认值'，当未指定默认值时，无此项
            # }
            'args': func_util.get_args(func)
        })

        return ENTRY_CACHE[fullname]

    def is_valid_route(self, func):
        source = inspect.getsource(func)
        lines = source.split('\n')
        for line in lines:
            if line.startswith('def '):
                # 已经查找到了函数定义部分了，说明没有找到
                return False

            if line.startswith('@route('):
                # 是 @route 装饰器行
                return True
        return False

    def get_route_map(self, route_path):
        # 命中
        hit_route = None
        for root_path in CONFIG_ROUTE:
            if route_path.startswith(root_path):
                hit_route = root_path, CONFIG_ROOT[APP_CONFIG_ROUTE][root_path]
                break

        if hit_route is None:
            return None

        # 将请求路径替换为指定的映射路径
        return ('%s%s' % (hit_route[1], route_path[len(hit_route[0]):])).strip('.')



from django.urls import path

from . import router

from .router import set_before_dispatch_handler

from .util import collector

urls = (
    [
        path('<str:entry>', router.dispatch),
        path('<str:entry>/<str:name>', router.dispatch)
    ],
    router.NAME,
    router.NAME
)
